# -*- coding: utf-8 -*-

"""Top-level package for parasweep."""

__author__ = """Eviatar Bach"""
__email__ = 'eviatarbach@protonmail.com'
__version__ = '2019.01'

from parasweep.sweep import run_sweep



"""Namers for generating simulation IDs."""
import math
from abc import ABC, abstractmethod


class Namer(ABC):
    """
    Abstract class for assigning simulation IDs to simulation.

    Only the `next` method has to be implemented.
    """

    def start(self, length):
        """
        Initialize naming.

        Parameters
        ----------
        length : int
            Indicates how many total simulations are in the sweep.

        """
        pass

    @abstractmethod
    def next(self, keys, param_set):
        """Generate simulation ID for a given parameter set.

        Parameters
        ----------
        keys : iterable
        param_set : iterable

        """
        pass


class SequentialNamer(Namer):
    """
    Name simulations with consecutive numbers and leading zeros.

    Examples
    --------
    >>> counter = SequentialNamer()
    >>> counter.start(length=11)
    >>> counter.next(['key1'], [['key_value1']])
    '00'
    >>> counter.next(['key2'], [['key_value2']])
    '01'

    """

    def __init__(self, zfill=None, start_at=0):
        """
        Initialize sequential naming object.

        Parameters
        ----------
        zfill : None or int, optional
            If provided, sets the width to which the name string is to be
            padded with zeros.
        start_at : int, optional
            Sets the integer to start at in the sequential naming.

        """
        self.zfill = zfill
        self.start_at = start_at

    def start(self, length):
        self.count = self.start_at - 1
        if self.zfill is None:
            # Compute how many digits are needed to represent all the numbers
            self.zfill = math.floor(math.log10(length - 1) + 1)
        self.length = length

    def next(self, keys, param_set):
        if self.count + 1 >= self.length + self.start_at:
            raise StopIteration
        self.count += 1
        return str(self.count).zfill(self.zfill)



"""Template engines for generating configuration files."""
from abc import ABC, abstractmethod
from string import Formatter


class Template(ABC):
    """
    Abstract base class for template engines.

    Subclasses should implement the `_load` and `render` methods. Make sure to
    implement errors for providing parameters not present in the template, and
    for using parameters in the template that are not provided.

    """

    def __init__(self, texts=None, paths=None):
        if (((paths is None) and (texts is None))
                or (not (paths is None) and not (texts is None))):
            raise ValueError('Exactly one of `paths` or `texts` must be '
                             'provided.')
        self.texts = texts
        self.paths = paths
        self._load()

    @abstractmethod
    def _load(self):
        pass

    @abstractmethod
    def render(self, params):
        """
        Render a configuration file with the template engine.

        Parameters
        ----------
        params : dict
            Dictionary with parameters and their values.

        """
        pass


class PythonFormatTemplate(Template):
    """Template engine using Python's string formatting."""

    def _load(self):
        if self.paths:
            self.templates = []
            for path in self.paths:
                with open(path, 'r') as template_file:
                    self.templates.append(template_file.read())
        else:
            self.templates = self.texts

    def render(self, params):
        keys = params.keys()
        unused_names = set(keys)
        rendered = []
        for template in self.templates:
            config_names = set([elem[1] for elem in
                                Formatter().parse(template) if elem[1]])
            unused_names -= config_names
            try:
                rendered.append(template.format(**params))
            except KeyError as key:
                raise NameError(f'The name {key} is used in the template but '
                                'not provided.')
        if unused_names:
            raise NameError(f'The names {unused_names} are not used in the '
                            'template.')
        return rendered


def _mako_template_names(template):
    """
    Return all the used identifiers in the Mako template.

    From Igonato's code at https://stackoverflow.com/a/23577289/622408.
    """
    from mako import lexer, codegen

    lexer = lexer.Lexer(template)
    node = lexer.parse()
    # ^ The node is the root element for the parse tree.
    # The tree contains all the data from a template
    # needed for the code generation process

    # Dummy compiler. _Identifiers class requires one
    # but only interested in the reserved_names field
    compiler = lambda: None
    compiler.reserved_names = set()

    identifiers = codegen._Identifiers(compiler, node)
    return identifiers.undeclared


class MakoTemplate(Template):
    """Template engine using Mako."""

    def _load(self):
        from mako.template import Template

        if self.paths:
            self.templates = []
            for path in self.paths:
                self.templates.append(Template(filename=path,
                                               input_encoding='utf-8',
                                               strict_undefined=True))
        else:
            self.templates = []
            for text in self.texts:
                self.templates.append(Template(text=text,
                                               input_encoding='utf-8',
                                               strict_undefined=True))

    def render(self, params):
        keys = params.keys()
        unused_names = set(keys)
        rendered = []
        for template in self.templates:
            config_names = _mako_template_names(template.source)
            unused_names -= config_names
            rendered.append(template.render_unicode(**params))
        if unused_names:
            raise NameError(f'The names {unused_names} are not used in the '
                            'template.')
        return rendered



"""
Test suite.

Tests will only work on UNIX-like systems. Some tests rely on drmaa, Mako, or
xarray being installed.
"""
from parasweep import run_sweep
from parasweep.namers import SequentialNamer
from parasweep.templates import PythonFormatTemplate

import unittest


class TestSweep(unittest.TestCase):
    def test_basic(self):
        out = open('out_test', 'w')
        out.close()
        run_sweep('cat {sim_id}.txt >> out_test', ['{sim_id}.txt'],
                  template_texts=['Hello {x:.2f}\n'],
                  sweep_parameters={'x': [1/3, 2/3, 3/3]})
        with open('out_test', 'r') as out:
            self.assertEqual(out.read(),
                             'Hello 0.33\nHello 0.67\nHello 1.00\n')

    def test_single(self):
        out = open('out_test', 'w')
        out.close()
        run_sweep('cat {sim_id}.txt >> out_test', ['{sim_id}.txt'],
                  template_texts=['Hello {x}, {y}\n'],
                  sweep_parameters={'x': [1], 'y': [2]})
        with open('out_test', 'r') as out:
            self.assertEqual(out.read(), 'Hello 1, 2\n')

    def test_multiple_templates(self):
        out = open('out_test', 'w')
        out.close()
        run_sweep('cat {sim_id}_1.txt {sim_id}_2.txt >> out_test',
                  ['{sim_id}_1.txt', '{sim_id}_2.txt'],
                  template_texts=['Hello {x:.2f}\n', 'Hello again {y}\n'],
                  sweep_parameters={'x': [1/3, 2/3, 3/3], 'y': [4]})
        with open('out_test', 'r') as out:
            self.assertEqual(out.read(),
                             'Hello 0.33\nHello again 4\nHello 0.67\n'
                             'Hello again 4\nHello 1.00\nHello again 4\n')

    def test_mako(self):
        from parasweep.templates import MakoTemplate
        out = open('out_test', 'w')
        out.close()
        run_sweep('cat {sim_id}.txt >> out_test', ['{sim_id}.txt'],
                  template_texts=['Hello ${x*10}\n'],
                  sweep_parameters={'x': [1, 2, 3]},
                  template_engine=MakoTemplate)
        with open('out_test', 'r') as out:
            self.assertEqual(out.read(),
                             'Hello 10\nHello 20\nHello 30\n')

    def test_drmaa(self):
        from parasweep.dispatchers import DRMAADispatcher

        out = open('out_test', 'w')
        out.close()
        run_sweep('cat {sim_id}.txt >> out_test', ['{sim_id}.txt'],
                  template_texts=['Hello {x}\n'],
                  sweep_parameters={'x': [1, 1, 1]}, wait=True,
                  dispatcher=DRMAADispatcher())
        with open('out_test', 'r') as out:
            self.assertEqual(out.read(), 'Hello 1\nHello 1\nHello 1\n')

    def test_delay(self):
        import time

        start_time = time.time()
        run_sweep('cat {sim_id}.txt', ['{sim_id}.txt'],
                  template_texts=['Hello {x}\n'],
                  sweep_parameters={'x': [1, 2]}, delay=2)

        self.assertGreater(time.time() - start_time, 4)

    def test_serial(self):
        import time
        from parasweep.dispatchers import DRMAADispatcher

        start_time = time.time()
        run_sweep('sleep 2', ['{sim_id}.txt'],
                  template_texts=['Hello {x}\n'],
                  sweep_parameters={'x': [1, 2]}, serial=True)

        self.assertGreater(time.time() - start_time, 4)

        start_time = time.time()
        run_sweep('sleep 2', ['{sim_id}.txt'],
                  template_texts=['Hello {x}\n'],
                  sweep_parameters={'x': [1, 2]}, dispatcher=DRMAADispatcher(),
                  serial=True)

        self.assertGreater(time.time() - start_time, 4)

    def test_wait(self):
        out = open('out_test', 'w')
        out.close()
        run_sweep('sleep 3; cat {sim_id}.txt >> out_test', ['{sim_id}.txt'],
                  template_texts=['Hello {x}\n'],
                  sweep_parameters={'x': [1, 1, 1]}, wait=True)
        with open('out_test', 'r') as out:
            self.assertEqual(out.read(), 'Hello 1\nHello 1\nHello 1\n')

    def test_errors(self):
        with self.assertRaises(TypeError) as context:
            run_sweep('cat {sim_id}.txt', '{sim_id}.txt',
                      template_texts='Hello {x} {z}\n',
                      sweep_parameters={'x': [1, 2, 3]})

        self.assertEqual('`config_paths` and `template_paths` or'
                         '`template_texts` must be a list.',
                         str(context.exception))

        with self.assertRaises(ValueError) as context:
            run_sweep('cat {sim_id}.txt', '{sim_id}.txt',
                      sweep_parameters={'x': [1, 2, 3]})

        self.assertEqual('Exactly one of `template_paths` or `template_texts` '
                         'must be provided.', str(context.exception))

    def test_param_mapping(self):
        param_array = run_sweep('cat {sim_id}.txt', ['{sim_id}.txt'],
                                template_texts=['Hello {x} {y} {z}\n'],
                                sweep_parameters={'x': [1, 2], 'y': [3, 4, 5],
                                                  'z': [6, 7, 8, 9]})
        self.assertEqual(param_array.coords.dims, ('x', 'y', 'z'))
        self.assertEqual(param_array.shape, (2, 3, 4))

    def test_parameter_sets(self):
        out = open('out_test', 'w')
        out.close()
        param_mapping = run_sweep('cat {sim_id}.txt >> out_test',
                                  ['{sim_id}.txt'],
                                  template_texts=['Hello {x}, {y}, {z}\n'],
                                  parameter_sets=[{'x': 2, 'y': 8, 'z': 5},
                                                  {'x': 1, 'y': -4, 'z': 9}],
                                  wait=True)
        with open('out_test', 'r') as out:
            self.assertEqual(out.read(), 'Hello 2, 8, 5\nHello 1, -4, 9\n')
        self.assertEqual(param_mapping, {'0': {'x': 2, 'y': 8, 'z': 5},
                                         '1': {'x': 1, 'y': -4, 'z': 9}})


class TestPythonTemplates(unittest.TestCase):
    def test_errors(self):
        with self.assertRaises(NameError) as context:
            run_sweep('cat {sim_id}.txt', ['{sim_id}.txt'],
                      template_texts=['Hello {x} {z}\n'],
                      sweep_parameters={'x': [1, 2, 3]})

        self.assertEqual("The name 'z' is used in the template but not "
                         "provided.",
                         str(context.exception))

        with self.assertRaises(NameError) as context:
            run_sweep('cat {sim_id}.txt', ['{sim_id}.txt'],
                      template_texts=['Hello {x}\n'],
                      sweep_parameters={'x': [1, 2, 3], 'y': [4]})

        self.assertEqual("The names {'y'} are not used in the template.",
                         str(context.exception))

        with self.assertRaises(ValueError) as context:
            PythonFormatTemplate()

        self.assertEqual('Exactly one of `paths` or `texts` must be '
                         'provided.', str(context.exception))

    def test_paths(self):
        out = open('out_test', 'w')
        out.close()

        with open('template_test.txt', 'w') as template_file:
            template_file.write('Hello {x}\n')

        run_sweep('cat {sim_id}.txt >> out_test', ['{sim_id}.txt'],
                  template_paths=['template_test.txt'],
                  sweep_parameters={'x': [1, 2, 3]})

        with open('out_test', 'r') as out:
            self.assertEqual(out.read(), 'Hello 1\nHello 2\nHello 3\n')


class TestMakoTemplates(unittest.TestCase):
    def test_errors(self):
        from parasweep.templates import MakoTemplate

        with self.assertRaises(NameError) as context:
            run_sweep('cat {sim_id}.txt', ['{sim_id}.txt'],
                      template_texts=['Hello ${x*10} ${z}\n'],
                      sweep_parameters={'x': [1, 2, 3]},
                      template_engine=MakoTemplate)

        self.assertEqual("'z' is not defined", str(context.exception))

        with self.assertRaises(NameError) as context:
            run_sweep('cat {sim_id}.txt', ['{sim_id}.txt'],
                      template_texts=['Hello ${x*10}\n'],
                      sweep_parameters={'x': [1, 2, 3], 'y': [4]},
                      template_engine=MakoTemplate)

        self.assertEqual("The names {'y'} are not used in the template.",
                         str(context.exception))

    def test_paths(self):
        from parasweep.templates import MakoTemplate

        out = open('out_test', 'w')
        out.close()

        with open('template_test.txt', 'w') as template_file:
            template_file.write('Hello ${x}\n')

        run_sweep('cat {sim_id}.txt >> out_test', ['{sim_id}.txt'],
                  template_paths=['template_test.txt'],
                  sweep_parameters={'x': [1, 2, 3]},
                  template_engine=MakoTemplate)

        with open('out_test', 'r') as out:
            self.assertEqual(out.read(), 'Hello 1\nHello 2\nHello 3\n')


class TestNamers(unittest.TestCase):
    def test_sequential(self):
        counter = SequentialNamer()
        counter.start(length=11)

        self.assertEqual(counter.next(['key1'], [['key_value1']]), '00')
        self.assertEqual(counter.next(['key2'], [['key_value2']]), '01')

        counter = SequentialNamer(zfill=3, start_at=3)
        counter.start(length=2)

        self.assertEqual(counter.next(['key1'], [['key_value1']]), '003')

        counter.next(['key2'], [['key_value2']])

        with self.assertRaises(StopIteration):
            counter.next(['key3'], [['key_value3']])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Main sweep functionality.
"""
from parasweep.namers import SequentialNamer
from parasweep.dispatchers import PythonSubprocessDispatcher
from parasweep.templates import PythonFormatTemplate

import itertools
import time
import operator
import datetime
import os
from functools import reduce


def run_sweep(command, config_paths, template_paths=None, template_texts=None,
              sweep_parameters={}, parameter_sets=[], sweep_id=None,
              naming=SequentialNamer(),
              dispatcher=PythonSubprocessDispatcher(),
              template_engine=PythonFormatTemplate, run=True, delay=0.0,
              serial=False, wait=False, cleanup=False, verbose=True,
              overwrite=True, mapping=True):
    r"""
    Run parameter sweeps.

    This function runs the program with the Cartesian product of the parameter
    ranges.

    Parameters
    ----------
    command : str
        The command to run. Must include '{sim_id}' indicating where the
        simulation ID is to be inserted.
    config_paths : list
        List of paths indicating where the configuration files should be saved
        after substitution of the parameters into the templates. Must be in the
        same order as `template_paths` or `template_texts`.
    Either
        **template_paths** (*list*)
            List of paths of templates to substitute parameters into. Must be
            in the same order as `config_paths`. Or,
        **template_texts** (*list*)
            List of template strings to substitute parameters into. Must be
            in the same order as `config_paths`.
    Either
        **sweep_parameters** (*dict*):
            Using this option will do a grid sweep (Cartesian product).
            Dictionary where the keys are the parameter names and the values
            are lists of parameter values to sweep over. E.g.,
            `{'x': [1, 2], 'y': [3, 4]}`. Or,
        **parameter_sets** (*list*):
            Using this option will run a sweep with the provided parameter
            sets. List of dictionaries where the keys are the parameter names
            and the values are the fixed parameter values. E.g.,
            `[{'x': 1, 'y': 2}, {'x': 3, 'y': 4}]`
    sweep_id : str, optional
        A name for the sweep. By default, the name is generated automatically
        from the date and time.
    naming : namers.Namer instance, optional
        A :class:`parasweep.namers.Namer` object that specifies how to assign
        simulation IDs. By default, assigns simulation IDs sequentially.
    dispatcher : dispatchers.Dispatcher instance, optional
        A :class:`parasweep.dispatchers.Dispatcher` object that specifies how
        to run the jobs. By default, uses Python's `subprocess` module.
    template_engine : templates.Template class, optional
        A :class:`parasweep.templates.Template` class that specifies the
        template engine to use. By default, uses Python format strings.
    run : bool, optional
        Whether to run the parameter sweep. True by default.
    delay : float, optional
        How many seconds to delay between running successive simulations.
        0.0 by default.
    serial : bool, optional
        Whether to run simulations serially, i.e., to wait for each simulation
        to complete before executing the next one. Enabling this turns off
        parallelism. False by default.
    wait : bool, optional
        Whether to wait for all simulations to complete before returning.
        False by default.
    cleanup : bool, optional
        Whether to delete configuration files after all the simulations are
        done. This will cause the command to wait on all processes before
        returning (as with the `wait` argument). False by default.
    verbose : bool, optional
        Whether to print some information about each simulation as it is
        launched. True by default.
    overwrite : bool, optional
        Whether to overwrite existing files when creating configuration files.
        If False, a `FileExistsError` will be raised when a configuration
        filename coincides with an existing one in the same directory. True by
        default.
    mapping : bool, optional
        Whether to return a mapping between the parameters to the simulation
        IDs. If the sweep is a grid sweep, an N-dimensional labelled array
        (using `xarray`) which maps the parameters to the simulation IDs will
        be returned. The array coordinates correspond to each sweep parameter,
        while the values contain the simulation IDs. This array will also be
        saved as a netCDF file with the name 'sim_ids_{sweep_id}.nc'. If
        instead specific parameter sets are provided (using the
        `parameter_sets` argument) then a dictionary mapping the simulation IDs
        to the parameter sets will be returned. True by default.

    Examples
    --------
    An example of the basic formatting that can be done with the Python
    formatting templates:

    >>> from parasweep import run_sweep
    >>> run_sweep('cat {sim_id}.txt', ['{sim_id}.txt'],
    ...           template_texts=['Hello {x:.2f}\n'],
    ...           sweep_parameters={'x': [1/3, 2/3, 3/3]},
    ...           verbose=False, mapping=False)
    Hello 0.33
    Hello 0.67
    Hello 1.00

    Mako templates provide functionality that is not available with Python
    formatting templates, being able to insert code within the template:

    >>> from parasweep.templates import MakoTemplate
    >>> run_sweep('cat {sim_id}.txt', ['{sim_id}.txt'],
    ...           template_texts=['Hello ${x*10}\n'],
    ...           sweep_parameters={'x': [1, 2, 3]}, verbose=False,
    ...           template_engine=MakoTemplate, mapping=False)
    Hello 10
    Hello 20
    Hello 30

    Multiple configuration files and their corresponding templates can be used:

    >>> run_sweep(command='cat {sim_id}_1.txt {sim_id}_2.txt',
    ...           config_paths=['{sim_id}_1.txt', '{sim_id}_2.txt'],
    ...           template_texts=['Hello {x:.2f}\n',
    ...                           'Hello again {y}\n'],
    ...           sweep_parameters={'x': [1/3, 2/3, 3/3], 'y': [4]},
    ...           verbose=False, mapping=False)
    Hello 0.33
    Hello again 4
    Hello 0.67
    Hello again 4
    Hello 1.00
    Hello again 4

    By default (if `mapping` is True), a mapping will be returned between
    the parameters and the simulation IDs, which facilitates postprocessing::

        >>> run_sweep('cat {sim_id}.txt >> out', ['{sim_id}.txt'],
        ...           template_texts=['Hello {x} {y} {z}\n'],
        ...           sweep_parameters={'x': [1, 2], 'y': [3, 4, 5],
        ...                             'z': [6, 7, 8, 9]})
        <xarray.DataArray 'sim_id' (x: 2, y: 3, z: 4)>
        array([[['0', '1', '2', '3'],
                ['4', '5', '6', '7'],
                ['8', '9', '10', '11']],

               [['12', '13', '14', '15'],
                ['16', '17', '18', '19'],
                ['20', '21', '22', '23']]], dtype='<U2')
        Coordinates:
          * x        (x) int64 1 2
          * y        (y) int64 3 4 5
          * z        (z) int64 6 7 8 9


    The default sweep is a Cartesian sweep, meaning that all the combinations
    of all the parameters are used (every member in the Cartesian product of
    the parameter values). Alternatively, specific parameter sets can be used:

    >>> mapping = run_sweep('cat {sim_id}.txt', ['{sim_id}.txt'],
    ...                     template_texts=['Hello {x}, {y}, {z}\n'],
    ...                     parameter_sets=[{'x': 2, 'y': 8, 'z': 5},
    ...                                     {'x': 1, 'y': -4, 'z': 9}],
    ...                     verbose=False)
    Hello 2, 8, 5
    Hello 1, -4, 9

    In the case that parameter sets are used, the parameter mapping is a
    dictionary like the following:

    >>> mapping
    {'0': {'x': 2, 'y': 8, 'z': 5}, '1': {'x': 1, 'y': -4, 'z': 9}}

    """
    if (((template_paths is None) and (template_texts is None))
            or (not (template_paths is None)
                and not (template_texts is None))):
        raise ValueError('Exactly one of `template_paths` or `template_texts` '
                         'must be provided.')

    if ((sweep_parameters and parameter_sets)
            or (not sweep_parameters and not parameter_sets)):
        raise ValueError('Exactly one of `sweep_parameters` or '
                         '`parameter_sets` must be provided.')

    if (isinstance(config_paths, str) or isinstance(template_paths, str)
            or isinstance(template_texts, str)):
        raise TypeError('`config_paths` and `template_paths` or'
                        '`template_texts` must be a list.')

    if not sweep_id:
        current_time = datetime.datetime.now()
        sweep_id = current_time.strftime('%Y-%m-%dT%H:%M:%S')

    if template_paths:
        config = template_engine(paths=template_paths)
    else:
        config = template_engine(texts=template_texts)

    if sweep_parameters:
        keys = list(sweep_parameters.keys())
        values = list(sweep_parameters.values())

        product = itertools.product(*values)
        lengths = [len(value) for value in values]
        sweep_length = reduce(operator.mul, lengths, 1)
    else:
        keys = parameter_sets[0].keys()

        product = parameter_sets
        sweep_length = len(parameter_sets)

    naming.start(length=sweep_length)

    sim_ids = []
    config_filenames = []

    if run:
        dispatcher.initialize_session()

    for param_set in product:
        if sweep_parameters:
            sweep_params = {}
            for index, value in enumerate(param_set):
                sweep_params[keys[index]] = value
        else:
            sweep_params = param_set

        sim_id = naming.next(keys, sweep_params.values())
        sim_ids.append(sim_id)

        rendered = config.render(sweep_params)
        for config_rendered, config_path in zip(rendered, config_paths):
            config_filename = config_path.format(sim_id=sim_id)
            config_filenames.append(config_filename)
            if not overwrite:
                if os.path.isfile(config_filename):
                    raise FileExistsError(f'{config_filename} exists, set '
                                          '`overwrite` to True to overwrite.')
            with open(config_filename, 'wb') as config_file:
                config_file.write(config_rendered.encode('utf-8', 'replace'))
        if run:
            if verbose:
                print(f'Running simulation {sim_id} with parameters:')
                print('\n'.join(f'{key}: {param}' for key, param
                                in sweep_params.items()))
            dispatcher.dispatch(command.format(sim_id=sim_id), serial)
            if delay:
                time.sleep(delay)

    if wait and run:
        dispatcher.wait_all()

    if cleanup and run:
        dispatcher.wait_all()
        for config_filename in config_filenames:
            os.remove(config_filename)

    if mapping and sweep_parameters:
        import xarray
        import numpy

        sim_ids_array = xarray.DataArray(numpy.reshape(numpy.array(sim_ids),
                                                       lengths),
                                         coords=values, dims=keys,
                                         name='sim_id')

        sim_ids_filename = f'sim_ids_{sweep_id}.nc'

        sim_ids_array.to_netcdf(sim_ids_filename)

        return sim_ids_array
    elif mapping and parameter_sets:
        return dict(zip(sim_ids, parameter_sets))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['xarray', 'numpy']

setup_requirements = []

test_requirements = []

setup(
    author="Eviatar Bach",
    author_email='eviatarbach@protonmail.com',
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="Utility for facilitating parallel parameter sweeps.",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='parasweep',
    name='parasweep',
    packages=find_packages(include=['parasweep']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    extras_require={'DRMAA support': ['drmaa'], 'Mako templates': ['mako']},
    url='https://github.com/eviatarbach/parasweep',
    version='2019.01',
    zip_safe=False,
)



"""Dispatchers for running parallel jobs."""
import subprocess
from abc import ABC, abstractmethod


class Dispatcher(ABC):
    """
    Abstract base class for dispatchers.

    Subclasses should implement the `initialize_session`, `dispatch`, and
    `wait_all` methods.

    """

    @abstractmethod
    def initialize_session(self):
        """Must be called before dispatching jobs."""
        pass

    @abstractmethod
    def dispatch(self, command, wait):
        """
        Dispatch a command using the dispatcher.

        Parameters
        ----------
        command : str
            The command to dispatch.
        wait : bool
            Whether to wait for the process to finish before returning.

        """
        pass

    @abstractmethod
    def wait_all(self):
        """Wait for the running/scheduled processes to finish, then return."""
        pass


class PythonSubprocessDispatcher(Dispatcher):
    def initialize_session(self):
        self.processes = []

    def dispatch(self, command, wait):
        process = subprocess.Popen(command, shell=True)
        self.processes.append(process)
        if wait:
            process.wait()

    def wait_all(self):
        for process in self.processes:
            process.wait()


class DRMAADispatcher(Dispatcher):
    """
    Dispatcher for DRMAA.

    Parameters
    ----------
    job_template : drmaa.JobTemplate instance, optional
        A job template containing settings for running the jobs with
        the job scheduler. Documentation for the different options is
        available in the Python drmaa package. Some options specific to each
        job scheduler, called the native specification, may have to be set
        using the `job_template.nativeSpecification` attribute, the options for
        which can be found in the job scheduler's DRMAA interface (e.g.,
        slurm-drmaa for Slurm and pbs-drmaa for PBS).

    Examples
    --------
    >>> import drmaa
    >>> jt = drmaa.JobTemplate(hardWallclockTimeLimit=60)
    >>> dispatcher = DRMAADispatcher(jt)

    """
    session = None

    def __init__(self, job_template=None):
        if job_template is None:
            import drmaa

            self.job_template = drmaa.JobTemplate()
        else:
            self.job_template = job_template

    def initialize_session(self):
        # Ensure there is only one active DRMAA session, otherwise it raises
        # an error
        if DRMAADispatcher.session is None:
            import drmaa

            self.session = drmaa.Session()
            self.session.initialize()
            DRMAADispatcher.session = self.session
        else:
            self.session = DRMAADispatcher.session

        self.jobids = []

    def dispatch(self, command, wait):
        self.job_template.remoteCommand = command

        jobid = self.session.runJob(self.job_template)
        self.jobids.append(jobid)
        if wait:
            self.session.wait(jobid)

    def wait_all(self):
        self.session.synchronize(self.jobids)
